import abc
import re
from enum import Enum
from typing import Type, TypeVar, Dict, Callable, Optional

import yaml
from cryptography.hazmat.primitives import serialization
from pydantic import BaseModel

from pki_tools.exceptions import (
    MissingInit,
    LoadError,
)

from loguru import logger

CryptoObject = TypeVar("CryptoObject")


class CryptoParser(BaseModel, abc.ABC):
    """
    CryptoParser is an abstract class used by all the types
    parsing cryptography objects into [pki_tools][pki_tools.types.certificate]
    pydantic classes.
    """

    _x509_obj: CryptoObject

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if "_x509_obj" not in kwargs:
            try:
                self._x509_obj = self._to_cryptography()
            except MissingInit:
                logger.trace("Can't create crypto object before init")

    @classmethod
    @abc.abstractmethod
    def from_cryptography(
        cls: Type["CryptoParser"], crypto_obj: CryptoObject
    ) -> "CryptoParser":
        """
        Parses a cryptography x509 object into a
        [CryptoParser][pki_tools.types.crypto_parser.CryptoParser]

        Arguments:
             crypto_obj: The cryptography object

        Returns:
            CryptoParser object
        """

    @abc.abstractmethod
    def _to_cryptography(self) -> CryptoObject:
        """
        Creates a x509 cryptography object from this class

        Returns: A x509 cryptography object
        """

    @abc.abstractmethod
    def _string_dict(self) -> Dict[str, str]:
        """
        Creates a dict representation of the object

        Returns: A dict containing all the keys in the CryptoParser
        """

    def __str__(self):
        return yaml.safe_dump(
            self._string_dict(),
            indent=2,
            default_flow_style=False,
            explicit_start=False,
            default_style="",
        )


class Encoding(Enum):
    """
    Describes the encoding used for writing/reading
    [InitCryptoParser][pki_tools.types.crypto_parser.InitCryptoParser]
    objects.

    Attributes:
        PEM: PEM string format
        DER: DER binary format
    """

    PEM = "pem_string"
    DER = "der_bytes"


class HelperFunc(BaseModel):
    func: Callable
    kwargs: Optional[Dict] = {}


class CryptoConfig(BaseModel):
    load_pem: HelperFunc
    load_der: HelperFunc
    pem_regexp: re.Pattern


class InitCryptoParser(CryptoParser, abc.ABC):
    """
    Extends the CryptoParser into an object that requires initialization
    before it can be used (while created as a
    [pki_tools][pki_tools.types.certificate] object and not loaded from
    cryptography). This can, for example, be a Certificate that needs
    to be signed with a KeyPair containing the private key.

    Attempt to e.g. dumping a certificate to a PEM string without using the
    sign (init) function first will result in a
    [MissingInit][pki_tools.exceptions.MissingInit] exception.

    Classes implementing the `InitCryptoParser` will also automatically get
    functions to load and write the objects with the supported
    [Encoding][pki_tools.types.crypto_parser.Encoding] as well as
    writing/reading the encoded content to/from files.
    """

    _init_func: str = "sign"

    @classmethod
    def from_pem_string(
        cls: Type["InitCryptoParser"], pem: str
    ) -> "InitCryptoParser":
        """
        Loads the object from a PEM string

        Arguments:
            pem: The PEM encoded object in string format

        Returns:
            A created object from the PEM
        """
        cfg = cls._crypto_config()

        if not isinstance(pem, str):
            logger.bind(pem=pem).debug("PEM parameter is not a string")
            raise LoadError()

        pem = re.sub(r"\n\s*", "\n", pem)

        if not re.match(cfg.pem_regexp, pem):
            logger.bind(pem=pem).debug("PEM parameter has invalid format")
            raise LoadError()

        func = cfg.load_pem.func
        kwargs = cfg.load_pem.kwargs
        crypto_obj = func(pem.encode(), **kwargs)
        return cls.from_cryptography(crypto_obj)

    @classmethod
    def from_der_bytes(
        cls: Type["InitCryptoParser"], der: bytes
    ) -> "InitCryptoParser":
        """
        Loads the object from DER bytes

        Arguments:
            der: The DER encoded object

        Returns:
            A created object from the DER bytes
        """
        cfg = cls._crypto_config()

        func = cfg.load_der.func
        kwargs = cfg.load_der.kwargs

        crypto_obj = func(der, **kwargs)
        return cls.from_cryptography(crypto_obj)

    @classmethod
    def from_file(
        cls, file_path: str, encoding: Encoding = Encoding.PEM
    ) -> "InitCryptoParser":
        """
        Reads a file containing one PEM into the object

        Arguments:
            file_path: The path to the file (can be relative the caller or
                absolute)
            encoding: The encoding to use for the dumped data
        Returns:
             The Certificate loaded from the specified file
        """
        mode = "r"
        if encoding == Encoding.DER:
            mode = "rb"

        with open(file_path, mode) as f:
            data = f.read()

        return getattr(cls, f"from_{encoding.value}")(data)

    @property
    def der_bytes(self) -> bytes:
        """
        Returns the DER bytes of the object

        Returns:
            The DER bytes.
        """
        return self._crypto_object.public_bytes(
            encoding=serialization.Encoding.DER
        )

    @property
    def pem_bytes(self) -> bytes:
        """
        Returns the PEM bytes of the object

        Returns:
            The PEM bytes.
        """
        return self._crypto_object.public_bytes(
            encoding=serialization.Encoding.PEM
        )

    @property
    def pem_string(self) -> str:
        """
        Returns:
            PEM decoded into a string
        """
        return self.pem_bytes.decode()

    def to_file(
        self, file_path: str, encoding: Encoding = Encoding.PEM
    ) -> None:
        """
        Saves the object with specified encoding to the specified file,
        creating it if it doesn't exist.

        Args:
            file_path: The path to the file (can be relative the caller or
                absolute)
            encoding: The encoding to use for the dumped data
        """
        mode = "w"
        if encoding == Encoding.DER:
            mode = "wb"
        with open(file_path, mode) as f:
            f.write(getattr(self, encoding.value))

    @classmethod
    @abc.abstractmethod
    def _crypto_config(cls) -> CryptoConfig:
        pass

    @property
    def _crypto_object(self) -> CryptoObject:
        if not hasattr(self, "_x509_obj") or self._x509_obj is None:
            init_func = f"{self.__class__.__name__}.{self._init_func}"
            raise MissingInit(f"Please use the {init_func} first")

        return self._x509_obj



from typing import Type, Optional, Dict
import re

import yaml
from cryptography import x509

from pki_tools.exceptions import MissingInit
from pki_tools.types.key_pair import (
    CryptoPublicKey,
    CryptoPrivateKey,
    CryptoKeyPair,
    Ed25519PrivateKey,
    Ed448PrivateKey,
)
from pki_tools.types.crypto_parser import (
    InitCryptoParser,
    CryptoConfig,
    HelperFunc,
)
from pki_tools.types.name import Name
from pki_tools.types.extensions import Extensions
from pki_tools.types.signature_algorithm import (
    SignatureAlgorithm,
)
from pki_tools.types.utils import _byte_to_hex

PEM_CSR_REGEX = re.compile(
    r"\s*-+BEGIN CERTIFICATE REQUEST-+[\w+/\s=]*-+END CERTIFICATE REQUEST-+\s*"
)


class CertificateSigningRequest(InitCryptoParser):
    """
    Represents a certificate signing request (CSR).

    Attributes:
        subject: The subject of the CSR.
        public_key: Public key associated with the CSR.
        extensions: Extensions associated with the CSR.
        attributes: Attributes of the CSR.
        signature_algorithm: Signature algorithm used while signing the CSR.

    --8<-- "docs/examples/csr.md"
    """

    subject: Name

    public_key: Optional[CryptoPublicKey] = None
    extensions: Optional[Extensions] = None
    attributes: Optional[Dict[str, bytes]] = None
    signature_algorithm: Optional[SignatureAlgorithm] = None

    _private_key: Optional[CryptoPrivateKey]

    @classmethod
    def from_cryptography(
        cls: Type["CertificateSigningRequest"],
        crypto_csr: x509.CertificateSigningRequest,
    ) -> "CertificateSigningRequest":
        """
        Create a CertificateSigningRequest object from a cryptography
        CertificateSigningRequest.

        Args:
            crypto_csr: Cryptography CertificateSigningRequest.

        Returns:
            Instance of CertificateSigningRequest.

        --8<-- "docs/examples/csr_from_cryptography.md"
        """
        attributes = {}
        for att in crypto_csr.attributes:
            attributes[att.oid.dotted_string] = att.value

        signature_algorithm = None
        if crypto_csr.signature_hash_algorithm:
            signature_algorithm = SignatureAlgorithm.from_cryptography(
                crypto_csr.signature_hash_algorithm,
                crypto_csr.signature_algorithm_parameters,
            )

        ret = cls(
            subject=Name.from_cryptography(crypto_csr.subject),
            extensions=Extensions.from_cryptography(crypto_csr.extensions),
            signature_algorithm=signature_algorithm,
            signature_value=_byte_to_hex(crypto_csr.signature),
            public_key=CryptoPublicKey.from_cryptography(
                crypto_csr.public_key()
            ),
            attributes=attributes,
            _x509_obj=crypto_csr,
        )
        ret._x509_obj = crypto_csr
        return ret

    @property
    def tbs_bytes(self) -> bytes:
        """
        Get the bytes to be signed of the CSR.

        Returns:
            TBS bytes of the CSR.
        """
        return self._crypto_object.tbs_certrequest_bytes

    def sign(
        self,
        key_pair: CryptoKeyPair,
        signature_algorithm: Optional[SignatureAlgorithm] = None,
    ):
        """
        Sign the CSR with the provided key pair and signature algorithm.

        Args:
            key_pair: Key pair with the private key to use while signing the
                CSR
            signature_algorithm: Signature algorithm to use for signing.
        """
        self._private_key = key_pair.private_key
        self.signature_algorithm = signature_algorithm
        self._x509_obj = self._to_cryptography()

    def _string_dict(self):
        ret = {
            "Certificate Signing Request": {
                "Subject": str(self.subject),
                "Public Key": self.public_key._string_dict(),
            }
        }
        if self.extensions:
            ret["Extensions"] = self.extensions._string_dict()
        if self.signature_algorithm is not None:
            signature_alg = self.signature_algorithm.algorithm.name.value
            ret["Signature Algorithm"] = signature_alg
        if self.attributes is not None:
            attributes = []

            for k in self.attributes:
                val = self.attributes[k]
                if k == "1.2.840.113549.1.9.14":
                    val = _byte_to_hex(val)
                elif k == "1.2.840.113549.1.9.7":
                    val = val.decode()
                attributes.append(f"{k}: {str(val)}")

            if attributes:
                ret["Certificate Signing Request"]["Attributes"] = attributes
        return ret

    def _to_cryptography(self) -> x509.CertificateSigningRequest:
        if not hasattr(self, "_private_key"):
            raise MissingInit(
                f"Please use CertificateSigningRequest."
                f"{self._init_func} function"
            )

        crypto_key = self._private_key._to_cryptography()

        builder = x509.CertificateSigningRequestBuilder()
        builder = builder.subject_name(self.subject._to_cryptography())

        if self.extensions is not None:
            for extension in self.extensions:
                builder = builder.add_extension(
                    extension._to_cryptography(), extension.critical
                )

        if self.attributes is not None:
            for attribute_oid, value in self.attributes.items():
                oid = x509.ObjectIdentifier(attribute_oid)
                builder = builder.add_attribute(oid, value)

        if isinstance(self._private_key, Ed448PrivateKey) or isinstance(
            self._private_key, Ed25519PrivateKey
        ):
            alg = None
        else:
            alg = self.signature_algorithm.algorithm._to_cryptography()

        return builder.sign(crypto_key, alg)

    def __str__(self) -> str:
        return yaml.safe_dump(
            self._string_dict(),
            indent=2,
            default_flow_style=False,
            explicit_start=True,
            default_style="",
        )

    @classmethod
    def _crypto_config(cls) -> CryptoConfig:
        return CryptoConfig(
            load_pem=HelperFunc(func=x509.load_pem_x509_csr),
            load_der=HelperFunc(func=x509.load_der_x509_csr),
            pem_regexp=PEM_CSR_REGEX,
        )



from loguru import logger

from pki_tools.types.extensions import UniformResourceIdentifier
from pki_tools.types.chain import Chain
from pki_tools.types.certificate import Certificate
from pki_tools.exceptions import (
    ExtensionMissing,
)
from pki_tools.types.crl import CertificateRevocationList


def _is_revoked(
    cert: Certificate,
    crl_issuer: Chain,
    crl_cache_seconds: int = 3600,
) -> bool:
    crl_issuer.check_chain()
    logger.debug("CRL issuer chain valid")

    log = logger.bind(
        cert=cert.pem_string,
        serial=cert.hex_serial,
    )

    if cert.extensions.crl_distribution_points is None:
        log.debug("CRL extension missing")
        raise ExtensionMissing()

    http_dist = False
    for dist_point in cert.extensions.crl_distribution_points:
        if dist_point.full_name is None:
            continue

        for full_name in dist_point.full_name:
            if not isinstance(full_name, UniformResourceIdentifier):
                logger.warning(
                    "CRL Distribution Point is not "
                    "UniformResourceIdentifier"
                )
                continue

            uri = full_name.value

            http_dist = True
            crl = CertificateRevocationList.from_uri(
                uri, cache_time_seconds=crl_cache_seconds
            )

            issuer = crl_issuer.get_issuer(crl)

            issuer.verify_signature(crl)
            logger.debug("CRL signature valid")

            if (r := crl.get_revoked(cert.serial_number)) is not None:
                log.bind(date=str(r.date)).debug("Certificate revoked")
                return True

    if not http_dist:
        log.debug("CRL missing URI")
        raise ExtensionMissing()

    log.debug("Certificate valid")
    return False



import base64
import hashlib
import re
from datetime import datetime
from enum import Enum
from typing import Type, Optional, Dict

from cryptography.x509 import ocsp
from cryptography.x509.ocsp import OCSPCertStatus
from loguru import logger

from pki_tools.types.extensions import Extensions
from pki_tools.types.certificate import Certificate
from pki_tools.exceptions import (
    MissingInit,
)
from pki_tools.types.key_pair import (
    CryptoPrivateKey,
    Ed25519PrivateKey,
    Ed448PrivateKey,
)
from pki_tools.types.crypto_parser import (
    InitCryptoParser,
    CryptoConfig,
    HelperFunc,
)
from pki_tools.types.signature_algorithm import (
    HashAlgorithm,
    SignatureAlgorithm,
)
from pki_tools.types.utils import _byte_to_hex


class OcspResponseStatus(Enum):
    """
    Enumeration of OCSP response statuses.
    """

    SUCCESSFUL = "SUCCESSFUL"
    MALFORMED_REQUEST = "MALFORMED_REQUEST"
    INTERNAL_ERROR = "INTERNAL_ERROR"
    TRY_LATER = "TRY_LATER"
    SIG_REQUIRED = "SIG_REQUIRED"
    UNAUTHORIZED = "UNAUTHORIZED"


class OcspCertificateStatus(Enum):
    """
    Enumeration of OCSP certificate statuses.
    """

    GOOD = "GOOD"
    REVOKED = "REVOKED"
    UNKNOWN = "UNKNOWN"


RESPONSE_REGEXP = re.compile(
    r"\s*-+BEGIN OCSP RESPONSE-+[\w+/\s=]*-+END OCSP RESPONSE-+\s*"
)


class OCSPResponse(InitCryptoParser):
    """
    Represents an OCSP response.

    Attributes:
        response_status: The OCSP response status.
        certificate_status: The OCSP certificate status.
        issuer_key_hash: The issuer key hash.
        revocation_time: The revocation time.

    --8<-- "docs/examples/ocsp_response.md"
    """

    response_status: OcspResponseStatus
    certificate_status: Optional[OcspCertificateStatus] = None
    issuer_key_hash: Optional[str] = None
    revocation_time: Optional[datetime] = None

    @classmethod
    def from_cryptography(
        cls: Type["OCSPResponse"], crypto_ocsp_response: ocsp.OCSPResponse
    ) -> "OCSPResponse":
        """
        Constructs an OCSPResponse object from a cryptography
        OCSPResponse object.

        Args:
            crypto_ocsp_response: The cryptography OCSPResponse object.

        Returns:
            OCSPResponse: The constructed OCSPResponse object.

        --8<-- "docs/examples/ocsp_response_from_cryptography.md"
        """
        response_status = crypto_ocsp_response.response_status

        ocsp_response_key_hash = None
        certificate_status = None
        revocation_time = None
        if response_status == ocsp.OCSPResponseStatus.SUCCESSFUL:
            certificate_status = crypto_ocsp_response.certificate_status.name
            try:
                ocsp_response_key_hash = _byte_to_hex(
                    crypto_ocsp_response.issuer_key_hash
                )
            except Exception as e:
                logger.bind(
                    exceptionType=type(e),
                    exception=str(e),
                    issuerHash=crypto_ocsp_response.issuer_key_hash,
                ).error("Couldn't convert issuer key hash to hex")
                raise

            if crypto_ocsp_response.revocation_time_utc is not None:
                revocation_time = crypto_ocsp_response.revocation_time_utc

        ret = cls(
            response_status=getattr(OcspResponseStatus, response_status.name),
            certificate_status=certificate_status,
            issuer_key_hash=ocsp_response_key_hash,
            revocation_time=revocation_time,
            _x509_obj=crypto_ocsp_response,
        )
        ret._x509_obj = crypto_ocsp_response
        return ret

    @property
    def tbs_bytes(self) -> bytes:
        """
        Returns the bytes to be singed of the OCSP response.

        Returns:
            bytes: The TBS bytes.
        """
        return self._crypto_object.tbs_response_bytes

    @property
    def pem_bytes(self) -> bytes:
        """
        Returns the PEM bytes of the object

        Returns:
            The PEM bytes.
        """
        ocsp_request_b64 = base64.b64encode(self.der_bytes).decode()

        # Create the PEM string with headers and footers
        return (
            "-----BEGIN OCSP RESPONSE-----\n"
            + "\n".join(
                [
                    ocsp_request_b64[i : i + 64]
                    for i in range(0, len(ocsp_request_b64), 64)
                ]
            )
            + "\n-----END OCSP RESPONSE-----\n"
        ).encode()

    def hash_with_alg(self, der_key: bytes) -> str:
        """
        Hashes a DER key bytes with the algorithm of the OCSP response.

        Args:
            der_key: The DER key.

        Returns:
            str: The hashed key.
        """

        hash_algorithm = hashlib.new(self._crypto_object.hash_algorithm.name)
        hash_algorithm.update(der_key)
        return hash_algorithm.hexdigest().upper()

    @property
    def is_successful(self) -> bool:
        """
        Checks if the OCSP response is successful.

        Returns:
            True if the response is successful, False otherwise.
        """
        return self.response_status == OcspResponseStatus.SUCCESSFUL

    @property
    def is_revoked(self) -> bool:
        """
        Checks if the certificate is revoked.

        Returns:
            True if the certificate is revoked, False otherwise.
        """
        return self.certificate_status == OcspCertificateStatus.REVOKED

    def sign(
        self,
        cert: Certificate,
        issuer: Certificate,
        response_algorithm: SignatureAlgorithm,
        private_key: CryptoPrivateKey,
        signature_algorithm: Optional[SignatureAlgorithm] = None,
    ):
        """
        Signs the OCSP response.

        Args:
            cert: The certificate.
            issuer: The issuer certificate.
            response_algorithm: The signature algorithm for the response.
            private_key: The private key to sign the response.
            signature_algorithm: The signature algorithm.
        """
        self._cert = cert
        self._issuer = issuer
        self._response_algorithm = response_algorithm
        self._private_key = private_key
        self._signature_algorithm = signature_algorithm
        self._x509_obj = self._to_cryptography()

    def _to_cryptography(self) -> ocsp.OCSPResponse:
        if not hasattr(self, "_private_key"):
            raise MissingInit(
                f"Please use OCSPResponse.{self._init_func} function"
            )

        builder = ocsp.OCSPResponseBuilder()
        cert_status = None
        if self.certificate_status is not None:
            cert_status = getattr(OCSPCertStatus, self.certificate_status.name)

        builder = builder.add_response(
            cert=self._cert._to_cryptography(),
            issuer=self._issuer._to_cryptography(),
            algorithm=self._response_algorithm.algorithm._to_cryptography(),
            cert_status=cert_status,
            this_update=datetime.now(),
            next_update=datetime.now(),
            revocation_reason=None,
            revocation_time=self.revocation_time,
        ).responder_id(
            ocsp.OCSPResponderEncoding.HASH, self._cert._to_cryptography()
        )

        if isinstance(self._private_key, Ed25519PrivateKey) or isinstance(
            self._private_key, Ed448PrivateKey
        ):
            alg = None
        else:
            alg = self._signature_algorithm.algorithm._to_cryptography()

        return builder.sign(
            self._private_key._to_cryptography(),
            alg,
        )

    def _string_dict(self) -> Dict[str, str]:
        ret = {
            "Response Status": self.response_status.value,
        }
        if self.certificate_status is not None:
            ret["Certificate Status"] = self.certificate_status.value
        if self.issuer_key_hash is not None:
            ret["Issuer Key Hash"] = self.issuer_key_hash
        if self.revocation_time is not None:
            ret["Revocation Time"] = self.revocation_time
        return ret

    @classmethod
    def _load_pem(cls, pem: bytes):
        pem_lines = pem.decode().strip().split("\n")
        base64_data = "".join(pem_lines[1:-1])
        der_bytes = base64.b64decode(base64_data)

        return ocsp.load_der_ocsp_response(der_bytes)

    @classmethod
    def _crypto_config(cls) -> CryptoConfig:
        return CryptoConfig(
            load_pem=HelperFunc(func=cls._load_pem),
            load_der=HelperFunc(func=ocsp.load_der_ocsp_response),
            pem_regexp=RESPONSE_REGEXP,
        )


REQUEST_REGEXP = re.compile(
    r"\s*-+BEGIN OCSP REQUEST-+[\w+/\s=]*-+END OCSP REQUEST-+\s*"
)


class OCSPRequest(InitCryptoParser):
    """
    Represents an OCSP request.

    Attributes:
        hash_algorithm: The hash algorithm.
        serial_number: The serial number.
        extensions: The extensions.

    --8<-- "docs/examples/ocsp_request.md"
    """

    hash_algorithm: HashAlgorithm

    serial_number: Optional[int] = None
    extensions: Optional[Extensions] = None

    _init_func = "create"

    @classmethod
    def from_cryptography(
        cls: Type["OCSPRequest"], crypto_obj: ocsp.OCSPRequest
    ) -> "OCSPRequest":
        """
        Constructs an OCSPRequest object from a cryptography OCSPRequest
        object.

        Args:
            crypto_obj: The cryptography OCSPRequest object.

        Returns:
            The constructed OCSPRequest object.

        --8<-- "docs/examples/ocsp_request_from_cryptography.md"
        """
        alg = HashAlgorithm.from_cryptography(crypto_obj.hash_algorithm)
        ret = cls(
            hash_algorithm=alg,
            serial_number=crypto_obj.serial_number,
            extensions=Extensions.from_cryptography(crypto_obj.extensions),
            _x590_obj=crypto_obj,
        )
        ret._x509_obj = crypto_obj
        return ret

    @property
    def pem_bytes(self) -> bytes:
        """
        Returns the PEM bytes of the object

        Returns:
            The PEM bytes.
        """
        ocsp_request_b64 = base64.b64encode(self.der_bytes).decode()

        return (
            "-----BEGIN OCSP REQUEST-----\n"
            + "\n".join(
                [
                    ocsp_request_b64[i : i + 64]
                    for i in range(0, len(ocsp_request_b64), 64)
                ]
            )
            + "\n-----END OCSP REQUEST-----\n"
        ).encode()

    @property
    def request_path(self) -> str:
        """
        The request path of the OCSP Response.

        Returns:
            The request path.
        """
        return base64.b64encode(self.der_bytes).decode()

    def create(self, cert: Certificate, issuer_cert: Certificate):
        """
        Creates an OCSP request.

        Args:
            cert: The certificate.
            issuer_cert: The issuer of the OCSP Response.
        """
        self._cert = cert
        self._issuer = issuer_cert
        self._x509_obj = self._to_cryptography()

    def _to_cryptography(self) -> ocsp.OCSPRequest:
        if not hasattr(self, "_cert"):
            raise MissingInit(
                f"Please use " f"OCSPRequest.{self._init_func} function"
            )

        builder = ocsp.OCSPRequestBuilder()
        builder = builder.add_certificate(
            self._cert._to_cryptography(),
            self._issuer._to_cryptography(),
            self.hash_algorithm._to_cryptography(),
        )
        return builder.build()

    def _string_dict(self) -> Dict:
        ret = self.hash_algorithm._string_dict()

        if self.serial_number is not None:
            ret["Serial Number"] = str(self.serial_number)
        if self.extensions is not None:
            ret["Extensions"] = self.extensions._string_dict()

        return ret

    @classmethod
    def _load_pem(cls, pem: bytes):
        pem_lines = pem.decode().strip().split("\n")
        base64_data = "".join(pem_lines[1:-1])
        der_bytes = base64.b64decode(base64_data)

        return ocsp.load_der_ocsp_request(der_bytes)

    @classmethod
    def _crypto_config(cls) -> CryptoConfig:
        return CryptoConfig(
            load_pem=HelperFunc(func=cls._load_pem),
            load_der=HelperFunc(func=ocsp.load_der_ocsp_request),
            pem_regexp=REQUEST_REGEXP,
        )



import abc
import importlib
import re
from enum import Enum
from typing import Dict, Type, Optional, get_args

import yaml
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives._serialization import (
    Encoding,
    PublicFormat,
    PrivateFormat,
)
from cryptography.hazmat.primitives.asymmetric import (
    dsa,
    ec,
    rsa,
    ed25519,
    ed448,
)
from cryptography.hazmat.primitives.asymmetric.types import (
    CertificateIssuerPrivateKeyTypes,
    CertificateIssuerPublicKeyTypes,
)

from pydantic import BaseModel

from pki_tools.types.crypto_parser import (
    CryptoConfig,
    HelperFunc,
    InitCryptoParser,
)
from pki_tools.types.signature_algorithm import PKCS1v15Padding
from pki_tools.types.utils import _byte_to_hex, _hex_to_byte


PUBLIC_KEY_REGEXP = re.compile(
    r"\s*-+BEGIN PUBLIC KEY-+[\w+/\s=]*-+END PUBLIC KEY-+\s*"
)
PRIVATE_KEY_REGEXP = re.compile(
    r"\s*-+BEGIN.+PRIVATE KEY-+[\w+/\s=]*-+END.+PRIVATE KEY-+\s*"
)


class CryptoPrivateKey(InitCryptoParser, abc.ABC):
    """
    Represents a cryptographic private key.
    """

    @classmethod
    def from_cryptography(
        cls: Type["CryptoPrivateKey"],
        key: CertificateIssuerPrivateKeyTypes,
    ) -> "CryptoPrivateKey":
        """
        Create a CryptoPrivateKey from a cryptography private key.

        Args:
            key: The cryptography private key.

        Returns:
            The CryptoPrivateKey object.
        """
        types = get_args(CertificateIssuerPrivateKeyTypes)

        name = None
        for check_type in types:
            if isinstance(key, check_type):
                name = check_type.__name__
                break

        return globals()[name].from_cryptography(key)

    @property
    def der_bytes(self) -> bytes:
        """
        Property to get the DER encoding of the public key.

        Returns:
            bytes: The DER encoded public key.
        """
        return self._crypto_object.private_bytes(
            encoding=Encoding.DER,
            format=PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

    @property
    def pem_bytes(self) -> bytes:
        """
        Property to get the PEM encoding of the private key.

        Returns:
            bytes: The PEM encoded private key.
        """
        kwargs = {
            "encoding": serialization.Encoding.PEM,
            "encryption_algorithm": serialization.NoEncryption(),
            "format": serialization.PrivateFormat.TraditionalOpenSSL,
        }

        if "Ed" in self.__class__.__name__:
            kwargs["format"] = serialization.PrivateFormat.PKCS8

        return self._crypto_object.private_bytes(**kwargs)

    @classmethod
    def _crypto_config(cls) -> CryptoConfig:
        return CryptoConfig(
            load_pem=HelperFunc(
                func=serialization.load_pem_private_key,
                kwargs={"password": None},
            ),
            load_der=HelperFunc(
                func=serialization.load_der_private_key,
                kwargs={"password": None},
            ),
            pem_regexp=PRIVATE_KEY_REGEXP,
        )


class CryptoPublicKey(InitCryptoParser, abc.ABC):
    """
    Represents a cryptographic public key.
    """

    @classmethod
    def from_cryptography(
        cls: Type["CryptoPublicKey"],
        key: CertificateIssuerPublicKeyTypes,
    ) -> "CryptoPublicKey":
        """
        Create a CryptoPublicKey from a cryptography public key.

        Args:
            key: The cryptography public key.

        Returns:
            The created CryptoPublicKey.
        """
        types = get_args(CertificateIssuerPublicKeyTypes)

        name = None
        for check_type in types:
            if isinstance(key, check_type):
                name = check_type.__name__
                break

        return globals()[name].from_cryptography(key)

    @property
    def der_bytes(self) -> bytes:
        """
        Property to get the DER encoding of the public key.

        Returns:
            bytes: The DER encoded public key.
        """
        return self._crypto_object.public_bytes(
            encoding=Encoding.DER,
            format=PublicFormat.SubjectPublicKeyInfo,
        )

    @property
    def ocsp_bytes(self) -> bytes:
        """
        The bytes used for the OCSP Response hash

        Returns:
            bytes: The bytes used for the OCSP Response hash
        """
        return self.der_bytes

    @property
    def pem_bytes(self) -> bytes:
        """
        Property to get the PEM encoding of the public key.

        Returns:
            bytes: The PEM encoded public key.
        """
        return self._crypto_object.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

    @abc.abstractmethod
    def verify(self, signed: InitCryptoParser) -> None:
        """
        Verifies the signature of a signed object.

        Args:
            signed: the signed object to verify.
        """

    @classmethod
    def _crypto_config(cls) -> CryptoConfig:
        return CryptoConfig(
            load_pem=HelperFunc(func=serialization.load_pem_public_key),
            load_der=HelperFunc(func=serialization.load_der_public_key),
            pem_regexp=PUBLIC_KEY_REGEXP,
        )


class CryptoKeyPair(BaseModel):
    """
    Represents a cryptographic key pair.

    Arguments:
        private_key: The private key
        public_key: The public key
    """

    private_key: CryptoPrivateKey
    public_key: CryptoPublicKey

    _init_func = "create"

    @classmethod
    @abc.abstractmethod
    def generate(cls: Type["CryptoKeyPair"], *args) -> "CryptoKeyPair":
        """
        Abstract method to generate a cryptographic key pair.

        Returns:
            The generated cryptographic key pair.
        """
        raise NotImplementedError

    def __str__(self):
        name = self.__class__.__name__
        d = {name: {}}
        if self.private_key is not None:
            d[name] = self.private_key._string_dict()
        else:
            d[name] = self.public_key._string_dict()
        return yaml.dump(d, indent=2)


class DSAPublicKey(CryptoPublicKey):
    """
    Represents a DSA cryptographic key pair.
    """

    key_size: int

    y: int
    p: int
    q: int
    g: int

    @classmethod
    def from_cryptography(
        cls: Type["DSAPublicKey"],
        key: dsa.DSAPublicKey,
    ) -> "DSAPublicKey":
        """
        Create a DSAKeyPair from a cryptography key.

        Args:
            key: The [cryptography.hazmat.primitives.asymmetric.dsa][]
                public key.

        Returns:
            DSAKeyPair: The DSA key pair.


        """
        public_numbers = key.public_numbers()
        param_numbers = key.parameters().parameter_numbers()

        ret = cls(
            y=public_numbers.y,
            p=param_numbers.p,
            q=param_numbers.q,
            g=param_numbers.g,
            key_size=key.key_size,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    @property
    def ocsp_bytes(self) -> bytes:
        """
        The bytes used for the OCSP Response hash

        Returns:
            bytes: The bytes used for the OCSP Response hash
        """
        return self._crypto_object.public_bytes(
            encoding=Encoding.DER,
            format=PublicFormat.SubjectPublicKeyInfo,
        )

    def verify(self, signed: InitCryptoParser):
        self._crypto_object.verify(
            signed._crypto_object.signature,
            signed.tbs_bytes,
            signed._crypto_object.signature_hash_algorithm,
        )

    def _to_cryptography(self) -> dsa.DSAPublicKey:
        public_numbers = dsa.DSAPublicNumbers(
            y=self.y,
            parameter_numbers=dsa.DSAParameterNumbers(
                p=self.p, q=self.q, g=self.g
            ),
        )
        return public_numbers.public_key()

    def _string_dict(self) -> Dict[str, str]:
        return {
            "key_size": str(self.key_size),
            "public_key_y": str(self.y),
            "prime_p": str(self.p),
            "subprime_q": str(self.q),
            "generator_g": str(self.g),
        }


class DSAPrivateKey(CryptoPrivateKey):
    """
    Represents a DSA cryptographic private key.
    """

    key_size: int
    x: int

    y: int
    p: int
    q: int
    g: int

    @classmethod
    def from_cryptography(
        cls: Type["DSAPrivateKey"],
        key: dsa.DSAPrivateKey,
    ) -> "DSAPrivateKey":
        """
        Create a DSAPrivateKey from a cryptography key.

        Args:
            key: The [cryptography.hazmat.primitives.asymmetric.dsa][]
                private key.

        Returns:
            DSAKeyPair: The DSA private key.
        """
        x = key.private_numbers().x

        public_numbers = key.public_key().public_numbers()
        param_numbers = key.parameters().parameter_numbers()

        ret = cls(
            x=x,
            y=public_numbers.y,
            p=param_numbers.p,
            q=param_numbers.q,
            g=param_numbers.g,
            key_size=key.key_size,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def _to_cryptography(self) -> dsa.DSAPrivateKey:
        public_numbers = dsa.DSAPublicNumbers(
            y=self.y,
            parameter_numbers=dsa.DSAParameterNumbers(
                p=self.p, q=self.q, g=self.g
            ),
        )

        private_numbers = dsa.DSAPrivateNumbers(
            x=self.x, public_numbers=public_numbers
        )
        return private_numbers.private_key()

    def _string_dict(self) -> Dict[str, str]:
        return {
            "key_size": str(self.key_size),
            "public_key_y": str(self.y),
            "prime_p": str(self.p),
            "subprime_q": str(self.q),
            "generator_g": str(self.g),
        }


class DSAKeyPair(CryptoKeyPair):
    """
    Describes a DSA Key Pair including the public and private DSA keys.

    --8<-- "docs/examples/key_pair_dsa.md"
    """

    @classmethod
    def generate(cls: Type["DSAKeyPair"], key_size: int) -> "DSAKeyPair":
        """
        Generate a DSA cryptographic key pair.

        Args:
            key_size: The size of the key.

        Returns:
            The generated DSA key pair.
        """
        crypto_private = dsa.generate_private_key(key_size=key_size)
        crypto_public = crypto_private.public_key()

        return cls(
            private_key=DSAPrivateKey.from_cryptography(crypto_private),
            public_key=DSAPublicKey.from_cryptography(crypto_public),
        )


class RSAPrivateKey(CryptoPrivateKey):
    """
    Represents an RSA cryptographic key pair.
    """

    key_size: int

    e: int
    n: int

    d: Optional[int]
    p: Optional[int]
    q: Optional[int]
    dmp1: Optional[int]
    dmq1: Optional[int]
    iqmp: Optional[int]

    @classmethod
    def from_cryptography(
        cls: Type["RSAPrivateKey"],
        key: rsa.RSAPrivateKey,
    ) -> "RSAPrivateKey":
        """
        Create an RSA private key from a cryptography key.

        Args:
            key: The cryptography private key.

        Returns:
            An RSAPrivateKey object
        """

        private_numbers = key.private_numbers()
        public_numbers = private_numbers.public_numbers
        d = private_numbers.d
        p = private_numbers.p
        q = private_numbers.q
        dmp1 = private_numbers.dmp1
        dmq1 = private_numbers.dmq1
        iqmp = private_numbers.iqmp

        ret = cls(
            e=public_numbers.e,
            n=public_numbers.n,
            d=d,
            p=p,
            q=q,
            dmp1=dmp1,
            dmq1=dmq1,
            iqmp=iqmp,
            key_size=key.key_size,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def _to_cryptography(self) -> rsa.RSAPrivateKey:
        public_numbers = rsa.RSAPublicNumbers(e=self.e, n=self.n)

        private_numbers = rsa.RSAPrivateNumbers(
            p=self.p,
            q=self.q,
            d=self.d,
            dmp1=self.dmp1,
            dmq1=self.dmq1,
            iqmp=self.iqmp,
            public_numbers=public_numbers,
        )
        return private_numbers.private_key()

    def _string_dict(self) -> Dict[str, str]:
        return {
            "key_size": str(self.key_size),
            "public_exponent_e": str(self.e),
            "modulus_n": str(self.n),
            "private_exponent_d": str(self.d),
            "prime_p": str(self.p),
            "prime_q": str(self.q),
            "dmp1": str(self.dmp1),
            "dmq1": str(self.dmq1),
            "iqmp": str(self.iqmp),
        }


class RSAPublicKey(CryptoPublicKey):
    """
    Represents an RSA cryptographic key pair.
    """

    key_size: int

    e: int
    n: int

    @classmethod
    def from_cryptography(
        cls: Type["RSAPublicKey"],
        key: rsa.RSAPublicKey,
    ) -> "RSAPublicKey":
        """
        Create an RSA public key from a cryptography public key.

        Args:
            key: The public cryptography key.

        Returns:
            The RSA public key
        """
        public_numbers = key.public_numbers()

        ret = cls(
            e=public_numbers.e,
            n=public_numbers.n,
            key_size=key.key_size,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    @property
    def der_bytes(self) -> bytes:
        """
        Property to get the DER encoding of the public key.

        Returns:
            bytes: The DER encoded public key.
        """
        return self._crypto_object.public_bytes(
            encoding=Encoding.DER,
            format=PublicFormat.PKCS1,
        )

    def verify(self, signed: InitCryptoParser) -> None:
        return self._crypto_object.verify(
            signed._crypto_object.signature,
            signed.tbs_bytes,
            PKCS1v15Padding()._to_cryptography(),
            signed._crypto_object.signature_hash_algorithm,
        )

    def _to_cryptography(self) -> rsa.RSAPublicKey:
        public_numbers = rsa.RSAPublicNumbers(e=self.e, n=self.n)

        return public_numbers.public_key()

    def _string_dict(self) -> Dict[str, str]:
        return {
            "key_size": str(self.key_size),
            "public_exponent_e": str(self.e),
            "modulus_n": str(self.n),
        }


class RSAKeyPair(CryptoKeyPair):
    """
    Describes an RSA Key Pair including the public and private RSA keys.

    --8<-- "docs/examples/key_pair_rsa.md"
    """

    @classmethod
    def generate(
        cls: Type["RSAKeyPair"], key_size: int = 2048, exponent: int = 65537
    ) -> "RSAKeyPair":
        """
        Generate an RSA cryptographic key pair.

        Args:
            key_size: The size of the key. Defaults to `2048`.
            exponent: The public exponent. Defaults to `65537`.

        Returns:
            The generated RSA key pair.
        """
        crypto_private = rsa.generate_private_key(
            public_exponent=exponent,
            key_size=key_size,
        )
        crypto_public = crypto_private.public_key()

        return cls(
            private_key=RSAPrivateKey.from_cryptography(crypto_private),
            public_key=RSAPublicKey.from_cryptography(crypto_public),
        )


EC_MODULE = importlib.import_module(
    "cryptography.hazmat.primitives.asymmetric.ec"
)


class EllipticCurveName(Enum):
    """
    Elliptic Curve Names
    """

    PRIME192V1 = "PRIME192V1"
    PRIME256V1 = "PRIME256V1"
    SECP192R1 = "SECP192R1"
    SECP224R1 = "SECP224R1"
    SECP256R1 = "SECP256R1"
    SECP384R1 = "SECP384R1"
    SECP521R1 = "SECP521R1"
    SECP256K1 = "SECP256K1"
    SECT163K1 = "SECT163K1"
    SECT233K1 = "SECT233K1"
    SECT283K1 = "SECT283K1"
    SECT409K1 = "SECT409K1"
    SECT571K1 = "SECT571K1"
    SECT163R2 = "SECT163R2"
    SECT233R1 = "SECT233R1"
    SECT283R1 = "SECT283R1"
    SECT409R1 = "SECT409R1"
    SECT571R1 = "SECT571R1"
    BRAINPOOLP256R1 = "BrainpoolP256R1"
    BRAINPOOLP384R1 = "BrainpoolP384R1"
    BRAINPOOLP512R1 = "BrainpoolP512R1"


class EllipticCurvePrivateKey(CryptoPrivateKey):
    """
    Represents an elliptic curve cryptographic private key.
    """

    curve_name: EllipticCurveName

    x: int
    y: int
    d: int

    @classmethod
    def from_cryptography(
        cls: Type["EllipticCurvePrivateKey"],
        key: ec.EllipticCurvePrivateKey,
    ) -> "EllipticCurvePrivateKey":
        """
        Create an EllipticCurvePrivateKey from a cryptography private key.

        Args:
            key: The cryptography private key.

        Returns:
            The EllipticCurvePrivateKey created object
        """

        private_numbers = key.private_numbers()
        public_key = key.public_key()
        public_numbers = public_key.public_numbers()
        curve_name = public_key.curve.name.upper()
        d = private_numbers.private_value

        ret = cls(
            x=public_numbers.x,
            y=public_numbers.y,
            curve_name=EllipticCurveName[curve_name],
            d=d,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def _to_cryptography(
        self,
    ) -> ec.EllipticCurvePrivateKey:
        curve = getattr(EC_MODULE, self.curve_name.value)()
        public_numbers = ec.EllipticCurvePublicNumbers(self.x, self.y, curve)

        private_numbers = ec.EllipticCurvePrivateNumbers(
            private_value=self.d, public_numbers=public_numbers
        )

        return private_numbers.private_key()

    def _string_dict(self) -> Dict[str, str]:
        return {
            "curve_name": self.curve_name.name,
            "x_coordinate": str(self.x),
            "y_coordinate": str(self.y),
        }


class EllipticCurvePublicKey(CryptoPublicKey):
    """
    Represents an elliptic curve cryptographic public key.
    """

    curve_name: EllipticCurveName

    x: int
    y: int

    @classmethod
    def from_cryptography(
        cls: Type["EllipticCurvePublicKey"],
        key: ec.EllipticCurvePublicKey,
    ) -> "EllipticCurvePublicKey":
        """
        Create an EllipticCurvePublicKey from a cryptography public key.

        Args:
            key: The cryptography public key.

        Returns:
            The created EllipticCurvePublicKey object
        """
        public_numbers = key.public_numbers()
        curve_name = key.curve.name.upper()

        ret = cls(
            x=public_numbers.x,
            y=public_numbers.y,
            curve_name=EllipticCurveName[curve_name],
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def verify(self, signed: InitCryptoParser):
        self._crypto_object.verify(
            signed._crypto_object.signature,
            signed.tbs_bytes,
            ec.ECDSA(signed._crypto_object.signature_hash_algorithm),
        )

    @property
    def ocsp_bytes(self) -> bytes:
        """
        The bytes used for the OCSP Response hash

        Returns:
            bytes: The bytes used for the OCSP Response hash
        """
        return self._crypto_object.public_bytes(
            encoding=Encoding.X962,
            format=PublicFormat.UncompressedPoint,
        )

    def _to_cryptography(
        self,
    ) -> ec.EllipticCurvePublicKey:
        curve = getattr(EC_MODULE, self.curve_name.value)()
        public_numbers = ec.EllipticCurvePublicNumbers(self.x, self.y, curve)
        return public_numbers.public_key()

    def _string_dict(self) -> Dict[str, str]:
        return {
            "curve_name": self.curve_name.name,
            "x_coordinate": str(self.x),
            "y_coordinate": str(self.y),
        }


class EllipticCurveKeyPair(CryptoKeyPair):
    """
    Describes an elliptic curve Key Pair including the public and private keys.

    --8<-- "docs/examples/key_pair_ec.md"
    """

    @classmethod
    def generate(
        cls: Type["EllipticCurveKeyPair"], curve_name: EllipticCurveName
    ) -> "EllipticCurveKeyPair":
        """
        Generate an elliptic curve cryptographic key pair.

        Args:
            curve_name: The name of the curve.

        Returns:
            The generated elliptic curve key pair.
        """
        crypto_private = ec.generate_private_key(
            curve=getattr(EC_MODULE, curve_name.value)()
        )
        crypto_public = crypto_private.public_key()

        return cls(
            private_key=EllipticCurvePrivateKey.from_cryptography(
                crypto_private
            ),
            public_key=EllipticCurvePublicKey.from_cryptography(crypto_public),
        )


class Ed448PrivateKey(CryptoPrivateKey):
    public_bytes: str
    private_bytes: str

    @classmethod
    def from_cryptography(
        cls: Type["Ed448PrivateKey"],
        key: ed448.Ed448PrivateKey,
    ) -> "Ed448PrivateKey":
        """
        Create an Ed25519KeyPair from a cryptography private key.

        Args:
            key: The cryptography private key.

        Returns:
            The created Ed448PrivateKey object.
        """

        public_key = key.public_key()
        private_bytes = _byte_to_hex(key.private_bytes_raw())

        ret = cls(
            public_bytes=_byte_to_hex(public_key.public_bytes_raw()),
            private_bytes=private_bytes,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def _to_cryptography(
        self,
    ) -> ed448.Ed448PrivateKey:
        return ed448.Ed448PrivateKey.from_private_bytes(
            _hex_to_byte(self.private_bytes)
        )

    def _string_dict(self) -> Dict[str, str]:
        return {
            "public_bytes": self.public_bytes,
        }


class Ed448PublicKey(CryptoPublicKey):
    public_bytes: str

    @classmethod
    def from_cryptography(
        cls: Type["Ed448PublicKey"],
        key: ed448.Ed448PublicKey,
    ) -> "Ed448PublicKey":
        """
        Create an Ed448PublicKey from a cryptography public key.

        Args:
            key: The cryptography public key.

        Returns:
            The Ed448PublicKey object.
        """
        ret = cls(
            public_bytes=_byte_to_hex(key.public_bytes_raw()),
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def verify(self, signed: InitCryptoParser):
        self._crypto_object.verify(
            signed._crypto_object.signature,
            signed.tbs_bytes,
        )

    def _to_cryptography(
        self,
    ) -> ed448.Ed448PublicKey:
        return ed448.Ed448PublicKey.from_public_bytes(
            _hex_to_byte(self.public_bytes)
        )

    def _string_dict(self) -> Dict[str, str]:
        return {
            "public_bytes": self.public_bytes,
        }


class Ed448KeyPair(CryptoKeyPair):
    """
    Represents an Ed448 cryptographic key pair.

    --8<-- "docs/examples/key_pair_ed448.md"
    """

    @classmethod
    def generate(cls: Type["Ed448KeyPair"]) -> "Ed448KeyPair":
        """
        Generate an Ed448 cryptographic key pair.

        Returns:
            The generated Ed448 key pair.
        """
        crypto_private = ed448.Ed448PrivateKey.generate()
        crypto_public = crypto_private.public_key()

        return cls(
            private_key=Ed448PrivateKey.from_cryptography(crypto_private),
            public_key=Ed448PublicKey.from_cryptography(crypto_public),
        )


class Ed25519PrivateKey(CryptoPrivateKey):
    private_bytes: str
    public_bytes: str

    @classmethod
    def from_cryptography(
        cls: Type["Ed25519PrivateKey"],
        key: ed25519.Ed25519PrivateKey,
    ) -> "Ed25519PrivateKey":
        """
        Create an Ed25519 private key from a cryptography private key.

        Args:
            key: The cryptography private key.

        Returns:
            The created Ed25519PrivateKey object.
        """

        public_key = key.public_key()
        private_bytes = _byte_to_hex(key.private_bytes_raw())

        ret = cls(
            public_bytes=_byte_to_hex(public_key.public_bytes_raw()),
            private_bytes=private_bytes,
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def _to_cryptography(
        self,
    ) -> ed25519.Ed25519PrivateKey:
        return ed25519.Ed25519PrivateKey.from_private_bytes(
            _hex_to_byte(self.private_bytes)
        )

    def _string_dict(self) -> Dict[str, str]:
        return {
            "public_bytes": self.public_bytes,
        }


class Ed25519PublicKey(CryptoPublicKey):
    public_bytes: str

    @classmethod
    def from_cryptography(
        cls: Type["Ed25519PublicKey"],
        key: ed25519.Ed25519PublicKey,
    ) -> "Ed25519PublicKey":
        """
        Create an Ed25519PublicKey from a cryptography public key.

        Args:
            key: The cryptography public key.

        Returns:
            The Ed25519PublicKey object.
        """
        ret = cls(
            public_bytes=_byte_to_hex(key.public_bytes_raw()),
            _x509_obj=key,
        )
        ret._x509_obj = key
        return ret

    def verify(self, signed: InitCryptoParser):
        self._crypto_object.verify(
            signed._crypto_object.signature,
            signed.tbs_bytes,
        )

    def _to_cryptography(
        self,
    ) -> ed25519.Ed25519PublicKey:
        return ed25519.Ed25519PublicKey.from_public_bytes(
            _hex_to_byte(self.public_bytes)
        )

    def _string_dict(self) -> Dict[str, str]:
        return {
            "public_bytes": self.public_bytes,
        }


class Ed25519KeyPair(CryptoKeyPair):
    """
    Represents an Ed25519 cryptographic key pair.

    --8<-- "docs/examples/key_pair_ed25519.md"
    """

    @classmethod
    def generate(cls: Type["Ed25519KeyPair"]) -> "Ed25519KeyPair":
        """
        Generate an Ed25519KeyPair cryptographic key pair.

        Returns:
            The generated Ed25519 key pair.
        """
        crypto_private = ed25519.Ed25519PrivateKey.generate()
        crypto_public = crypto_private.public_key()

        return cls(
            private_key=Ed25519PrivateKey.from_cryptography(crypto_private),
            public_key=Ed25519PublicKey.from_cryptography(crypto_public),
        )
